/*
 *  drivers/cpufreq/cpufreq_cs_freqtable.c
 *
 *  Copyright (C)  2001 Russell King
 *            (C)  2003 Venkatesh Pallipadi <venkatesh.pallipadi@intel.com>.
 *                      Jun Nakajima <jun.nakajima@intel.com>
 *            (C)  2009 Alexander Clouter <alex@digriz.org.uk>
 *            (C)  2014 Lantiq; derived from conservative governor
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/cpufreq.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/kernel_stat.h>
#include <linux/kobject.h>
#include <linux/module.h>
#include <linux/mutex.h>
#include <linux/notifier.h>
#include <linux/percpu-defs.h>
#include <linux/ratelimit.h>
#include <linux/slab.h>
#include <linux/sysfs.h>
#include <linux/types.h>
#include <cpufreq/ltq_cpufreq.h>

#undef AVM_LOAD_LOG
#if defined(AVM_LOAD_LOG)
#include <linux/simple_proc.h>
#endif

#include "cpufreq_governor.h"

/* lantiqgov governor macros */
/*
#define DEF_FREQUENCY_UP_THRESHOLD	(80)
#define DEF_FREQUENCY_DOWN_THRESHOLD	(20)
#define DEF_SAMPLING_DOWN_FACTOR	(1)
*/
#define MAX_SAMPLING_DOWN_FACTOR	(20)
#define DEF_FREQUENCY_STEP		(5) /*not used*/
#define DOWN				(0)
#define UP				(1)

static DEFINE_PER_CPU(struct cs_cpu_dbs_info_s, cs_cpu_dbs_info);
extern int ltq_get_subsys_busy(void);
static DEFINE_PER_CPU(unsigned int, cpufreq_stat_stepdown);

static  unsigned int get_freq_target(struct cs_dbs_tuners *cs_tuners,
				 struct cpufreq_policy *policy, int dir)
{
	struct cpufreq_frequency_table *table;
	int i, found = -1;
	unsigned int freq;

	table = cpufreq_frequency_get_table(policy->cpu);

	if (dir == DOWN) {
		/* fetch next lower frequency */
		for (i = 0; (table[i].frequency != CPUFREQ_TABLE_END); i++) {
			freq = table[i].frequency;
			if (freq == CPUFREQ_ENTRY_INVALID) {
				pr_debug("tab entry %u invalid, skip\n", i);
				continue;
			}
			pr_debug("tab entry %u: %u kHz, %u index\n", i, freq,
					 table[i].index);
			if (freq >= policy->cur)
				continue;
			/*smaller frequency found*/
			found = 1;
			break;
		}
	} else {
		/* fetch next higher frequency */
		/*first search end of list*/
		for (i = 0; (table[i].frequency != CPUFREQ_TABLE_END); i++)
			;
		i--; /*leave CPUFREQ_TABLE_END*/
		for (; i >= 0; i--) {
			freq = table[i].frequency;
			if (freq == CPUFREQ_ENTRY_INVALID) {
				pr_debug("tab entry %u invalid, skip\n", i);
				continue;
			}
			pr_debug("tab entry %u: %u kHz, %u index\n", i, freq,
					 table[i].index);
			if (freq <= policy->cur)
				continue;
			/*higher frequency found*/
			found = 1;
			break;
		}
	}
	if (found < 0)
		return -EINVAL;
	else
		return freq;
}

static const atomic_t *kick_locked_ptr;
static atomic_t ltq_kick_disable_check = ATOMIC_INIT(0);

static inline int kick_locked(void)
{
	if (!kick_locked_ptr)
		return 0; /* ptr not initialized - not locked */

	if (unlikely(atomic_read(&ltq_kick_disable_check))) {
		pr_debug_ratelimited("%s: in fake-high-load mode\n", __func__);
		return 0; /* disable check -> not locked */
	}

	if (unlikely(atomic_read(kick_locked_ptr))) {
		pr_debug_ratelimited("%s: still kick-locked\n", __func__);
		return 1; /* lock variable was set from kick caller -> locked */
	}

	return 0;
}

#if defined(AVM_LOAD_LOG)
#define LOAD_LOG_SIZE 256
struct load_log_entry{
	u32 cpu;
	u32 load;
	u64 jiffies;
	struct timespec uptime;
};

static struct load_log_entry log_load_entries[LOAD_LOG_SIZE];
atomic_t load_log_pos = ATOMIC_INIT(0);

void log_load(u32 cpu, u32 load){
	u32 mypos = atomic_read(&load_log_pos);
	atomic_set(&load_log_pos, (mypos+1)%LOAD_LOG_SIZE);

	log_load_entries[mypos].cpu = cpu;
	log_load_entries[mypos].load = load;
	log_load_entries[mypos].jiffies = get_jiffies_64();

	do_posix_clock_monotonic_gettime(&log_load_entries[mypos].uptime);
	monotonic_to_bootbased(&log_load_entries[mypos].uptime);
}

void print_load_log(struct seq_file *m, void *data __maybe_unused){
	u32 i;
	u32 end_pos = atomic_read(&load_log_pos);

	for ( i = (end_pos-1)%LOAD_LOG_SIZE; i != end_pos; i =
	      (i-1)%LOAD_LOG_SIZE){

		seq_printf(m, "%03lu.%03lu <%u>: %u\n",

			(unsigned long) log_load_entries[i].uptime.tv_sec,
			(log_load_entries[i].uptime.tv_nsec / (NSEC_PER_SEC / 1000)),
			log_load_entries[i].cpu,
			log_load_entries[i].load);
	}
}
#endif
/*
 * Every sampling_rate, we check, if current idle time is less than 20%
 * (default), then we try to increase frequency. Every sampling_rate *
 * sampling_down_factor, we check, if current idle time is more than 80%
 * (default), then we try to decrease frequency
 *
 * Any frequency increase takes it to the maximum frequency. Frequency
 * reduction happens step by step as defined in the frequency table.
 */
static void cs_check_cpu(int cpu, unsigned int load)
{
	struct cs_cpu_dbs_info_s *dbs_info = &per_cpu(cs_cpu_dbs_info, cpu);
	struct cpufreq_policy *policy = dbs_info->cdbs.cur_policy;
	struct dbs_data *dbs_data = policy->governor_data;
	struct cs_dbs_tuners *cs_tuners = dbs_data->tuners;
	/*int freq;*/
	
	//notify of no change so cpufreq stats can continuously measure
	//statistics even when frequency does not change
    struct cpufreq_freqs freqs;
    freqs.cpu=cpu;
    freqs.old=policy->cur;
    freqs.new=policy->cur;
    
    //however only do this every tenth time
    if( per_cpu(cpufreq_stat_stepdown, cpu) < 9 ) {
        per_cpu(cpufreq_stat_stepdown, cpu)++;
    } else {
        per_cpu(cpufreq_stat_stepdown, cpu)=0;
        cpufreq_notify_transition(policy, &freqs, CPUFREQ_NOCHANGE);
    }

#if defined(AVM_LOAD_LOG)
	log_load(cpu, load);
#endif

	if (kick_locked())
		return;

	/*
	 * break out if we 'cannot' reduce the speed as the user might
	 * want freq_step to be zero
	 */
	if (cs_tuners->freq_step == 0)
		return;

	/* Check for frequency increase */
	if (load > cs_tuners->up_threshold) {
		dbs_info->down_skip = 0;

		/* if we are already at full speed then break out early */
		if ((dbs_info->requested_freq == policy->max) &&
			(policy->cur == policy->max))
			return;


		/*
		freq = get_freq_target(cs_tuners, policy, UP);
		if (freq < 0)
		        return;
		*/

		dbs_info->requested_freq = policy->max; /*freq;*/
		__cpufreq_driver_target(policy, dbs_info->requested_freq,
			CPUFREQ_RELATION_H);
		return;
	}

	/* if sampling_down_factor is active break out early */
	if (++dbs_info->down_skip < cs_tuners->sampling_down_factor)
		return;
	dbs_info->down_skip = 0;

	/* Check for frequency decrease */
	if ((load < cs_tuners->down_threshold) &&
						(0 == ltq_get_subsys_busy())) {
		/*
		 * if we cannot reduce the frequency anymore, break out early
		 */
		if (policy->cur == policy->min)
			return;

		dbs_info->requested_freq =
				get_freq_target(cs_tuners, policy, DOWN);
/*
		requested_freq is unsigned int therefore this check makes
		if (dbs_info->requested_freq < 0)
			return;

*/
		__cpufreq_driver_target(policy, dbs_info->requested_freq,
				CPUFREQ_RELATION_L);
		return;
	}
}

static void cs_dbs_timer(struct work_struct *work)
{
	struct cs_cpu_dbs_info_s *dbs_info = container_of(work,
			struct cs_cpu_dbs_info_s, cdbs.work.work);
	unsigned int cpu = dbs_info->cdbs.cur_policy->cpu;
	struct cs_cpu_dbs_info_s *core_dbs_info = &per_cpu(cs_cpu_dbs_info,
			cpu);
	struct dbs_data *dbs_data = dbs_info->cdbs.cur_policy->governor_data;
	struct cs_dbs_tuners *cs_tuners = dbs_data->tuners;
	int delay = delay_for_sampling_rate(cs_tuners->sampling_rate);
	bool modify_all = true;

	mutex_lock(&core_dbs_info->cdbs.timer_mutex);
	if (!need_load_eval(&core_dbs_info->cdbs, cs_tuners->sampling_rate))
		modify_all = false;
	else
		dbs_check_cpu(dbs_data, cpu);

	gov_queue_work(dbs_data, dbs_info->cdbs.cur_policy, delay, modify_all);
	mutex_unlock(&core_dbs_info->cdbs.timer_mutex);
}

static int dbs_cpufreq_notifier(struct notifier_block *nb, unsigned long val,
		void *data)
{
	struct cpufreq_freqs *freq = data;
	struct cs_cpu_dbs_info_s *dbs_info =
					&per_cpu(cs_cpu_dbs_info, freq->cpu);
	struct cpufreq_policy *policy;

	if (!dbs_info->enable)
		return 0;

	policy = dbs_info->cdbs.cur_policy;

	/*
	 * we only care if our internally tracked freq moves outside the valid
	 * ranges of frequency available to us otherwise we do not change it
	*/
	if (dbs_info->requested_freq > policy->max
			|| dbs_info->requested_freq < policy->min)
		dbs_info->requested_freq = freq->new;

	return 0;
}

/************************** sysfs interface ************************/
static struct common_dbs_data cs_dbs_cdata;

static ssize_t store_sampling_down_factor(struct dbs_data *dbs_data,
		const char *buf, size_t count)
{
	struct cs_dbs_tuners *cs_tuners = dbs_data->tuners;
	unsigned int input;
	int ret;
	ret = sscanf(buf, "%u", &input);

	if (ret != 1 || input > MAX_SAMPLING_DOWN_FACTOR || input < 1)
		return -EINVAL;

	cs_tuners->sampling_down_factor = input;
	return count;
}

static ssize_t store_sampling_rate(struct dbs_data *dbs_data, const char *buf,
		size_t count)
{
	struct cs_dbs_tuners *cs_tuners = dbs_data->tuners;
	unsigned int input;
	int ret;
	ret = sscanf(buf, "%u", &input);

	if (ret != 1)
		return -EINVAL;

	cs_tuners->sampling_rate = max(input, dbs_data->min_sampling_rate);
	return count;
}

static ssize_t store_up_threshold(struct dbs_data *dbs_data, const char *buf,
		size_t count)
{
	struct cs_dbs_tuners *cs_tuners = dbs_data->tuners;
	unsigned int input;
	int ret;
	ret = sscanf(buf, "%u", &input);

	if (ret != 1 || input > 100 || input <= cs_tuners->down_threshold)
		return -EINVAL;

	cs_tuners->up_threshold = input;
	return count;
}

static ssize_t store_down_threshold(struct dbs_data *dbs_data, const char *buf,
		size_t count)
{
	struct cs_dbs_tuners *cs_tuners = dbs_data->tuners;
	unsigned int input;
	int ret;
	ret = sscanf(buf, "%u", &input);

	/* cannot be lower than 11 otherwise freq will not fall */
	if (ret != 1 || input < 11 || input > 100 ||
			input >= cs_tuners->up_threshold)
		return -EINVAL;

	cs_tuners->down_threshold = input;
	return count;
}

static ssize_t store_ignore_nice_load(struct dbs_data *dbs_data,
		const char *buf, size_t count)
{
	struct cs_dbs_tuners *cs_tuners = dbs_data->tuners;
	unsigned int input, j;
	int ret;

	ret = sscanf(buf, "%u", &input);
	if (ret != 1)
		return -EINVAL;

	if (input > 1)
		input = 1;

	if (input == cs_tuners->ignore_nice_load) /* nothing to do */
		return count;

	cs_tuners->ignore_nice_load = input;

	/* we need to re-evaluate prev_cpu_idle */
	for_each_online_cpu(j) {
		struct cs_cpu_dbs_info_s *dbs_info;
		dbs_info = &per_cpu(cs_cpu_dbs_info, j);
		dbs_info->cdbs.prev_cpu_idle = get_cpu_idle_time(j,
					&dbs_info->cdbs.prev_cpu_wall, 0);
		if (cs_tuners->ignore_nice_load)
			dbs_info->cdbs.prev_cpu_nice =
				kcpustat_cpu(j).cpustat[CPUTIME_NICE];
	}
	return count;
}

static ssize_t store_freq_step(struct dbs_data *dbs_data, const char *buf,
		size_t count)
{
	struct cs_dbs_tuners *cs_tuners = dbs_data->tuners;
	unsigned int input;
	int ret;
	ret = sscanf(buf, "%u", &input);

	if (ret != 1)
		return -EINVAL;

	if (input > 100)
		input = 100;

	/*
	 * no need to test here if freq_step is zero as the user might actually
	 * want this, they would be crazy though :)
	 */
	cs_tuners->freq_step = input;
	return count;
}

static int ltq_cpufreq_kick(unsigned cpu, const atomic_t *kick)
{
	WARN_ON(kick_locked_ptr && kick_locked_ptr != kick);
	kick_locked_ptr = kick;

	atomic_inc(&ltq_kick_disable_check);

	pr_debug("%s: Faking high-load on cpu %u\n", __func__, cpu);
	cs_check_cpu(cpu, 100);

	atomic_dec(&ltq_kick_disable_check);
	pr_debug("%s: Disabled load sensitivity until kick will be disabled\n",
		 __func__);

	return 0;
}

show_store_one(cs, sampling_rate);
show_store_one(cs, sampling_down_factor);
show_store_one(cs, up_threshold);
show_store_one(cs, down_threshold);
show_store_one(cs, ignore_nice_load);
show_store_one(cs, freq_step);
declare_show_sampling_rate_min(cs);

gov_sys_pol_attr_rw(sampling_rate);
gov_sys_pol_attr_rw(sampling_down_factor);
gov_sys_pol_attr_rw(up_threshold);
gov_sys_pol_attr_rw(down_threshold);
gov_sys_pol_attr_rw(ignore_nice_load);
gov_sys_pol_attr_rw(freq_step);
gov_sys_pol_attr_ro(sampling_rate_min);

static struct attribute *dbs_attributes_gov_sys[] = {
	&sampling_rate_min_gov_sys.attr,
	&sampling_rate_gov_sys.attr,
	&sampling_down_factor_gov_sys.attr,
	&up_threshold_gov_sys.attr,
	&down_threshold_gov_sys.attr,
	&ignore_nice_load_gov_sys.attr,
	&freq_step_gov_sys.attr,
	NULL
};

static struct attribute_group cs_attr_group_gov_sys = {
	.attrs = dbs_attributes_gov_sys,
	.name = "lantiqgov",
};

static struct attribute *dbs_attributes_gov_pol[] = {
	&sampling_rate_min_gov_pol.attr,
	&sampling_rate_gov_pol.attr,
	&sampling_down_factor_gov_pol.attr,
	&up_threshold_gov_pol.attr,
	&down_threshold_gov_pol.attr,
	&ignore_nice_load_gov_pol.attr,
	&freq_step_gov_pol.attr,
	NULL
};

static struct attribute_group cs_attr_group_gov_pol = {
	.attrs = dbs_attributes_gov_pol,
	.name = "lantiqgov",
};

/************************** sysfs end ************************/

static int cs_init(struct dbs_data *dbs_data)
{
	struct ltq_cpufreq *ltq_cpufreq_obj = ltq_cpufreq_get();
	struct cs_dbs_tuners *tuners;

	tuners = kzalloc(sizeof(struct cs_dbs_tuners), GFP_KERNEL);
	if (!tuners) {
		pr_err("%s: kzalloc failed\n", __func__);
		return -ENOMEM;
	}

	tuners->up_threshold = ltq_cpufreq_obj->frequency_up_threshold;
	tuners->down_threshold = ltq_cpufreq_obj->frequency_down_threshold;
	tuners->sampling_down_factor = ltq_cpufreq_obj->sampling_down_factor;
	tuners->ignore_nice_load = 0;
	tuners->freq_step = DEF_FREQUENCY_STEP;

	dbs_data->tuners = tuners;
	dbs_data->min_sampling_rate = MIN_SAMPLING_RATE_RATIO *
		jiffies_to_usecs(10);
	mutex_init(&dbs_data->mutex);
	return 0;
}

static void cs_exit(struct dbs_data *dbs_data)
{
	kfree(dbs_data->tuners);
}

define_get_cpu_dbs_routines(cs_cpu_dbs_info);

static struct notifier_block cs_cpufreq_notifier_block = {
	.notifier_call = dbs_cpufreq_notifier,
};

static struct cs_ops cs_ops = {
	.notifier_block = &cs_cpufreq_notifier_block,
};

static struct common_dbs_data cs_dbs_cdata = {
	.governor = GOV_CONSERVATIVE,
	.attr_group_gov_sys = &cs_attr_group_gov_sys,
	.attr_group_gov_pol = &cs_attr_group_gov_pol,
	.get_cpu_cdbs = get_cpu_cdbs,
	.get_cpu_dbs_info_s = get_cpu_dbs_info_s,
	.gov_dbs_timer = cs_dbs_timer,
	.gov_check_cpu = cs_check_cpu,
	.gov_ops = &cs_ops,
	.kick = ltq_cpufreq_kick,
	.init = cs_init,
	.exit = cs_exit,
};

static int cs_cpufreq_governor_dbs(struct cpufreq_policy *policy,
				   unsigned int event)
{
	return cpufreq_governor_dbs(policy, &cs_dbs_cdata, event);
}

#ifndef CONFIG_CPU_FREQ_DEFAULT_GOV_LANTIQGOV
static
#endif
struct cpufreq_governor cpufreq_gov_lantiqgov = {
	.name			= "lantiqgov",
	.governor		= cs_cpufreq_governor_dbs,
	.max_transition_latency	= TRANSITION_LATENCY_LIMIT,
	.owner			= THIS_MODULE,
};

static int __init cpufreq_gov_dbs_init(void)
{
    
	unsigned int cpu;
	for_each_online_cpu(cpu) {
	    per_cpu(cpufreq_stat_stepdown, cpu) = 0;
	}

#if defined(AVM_LOAD_LOG)
	add_simple_proc_file("avm_load_log", NULL, print_load_log, NULL);
#endif
	return cpufreq_register_governor(&cpufreq_gov_lantiqgov);
}

static void __exit cpufreq_gov_dbs_exit(void)
{
	cpufreq_unregister_governor(&cpufreq_gov_lantiqgov);
}

MODULE_AUTHOR("Thomas Bartholomae <t.bartholomae@lantiq.com>");
MODULE_DESCRIPTION("'cpufreq_lantiqgov' - A dynamic cpufreq governor based\n"
		"on conservative governor but frequency steps are given\n"
		"by frequency table");
MODULE_LICENSE("GPL");

#ifdef CONFIG_CPU_FREQ_DEFAULT_GOV_LANTIQGOV
fs_initcall(cpufreq_gov_dbs_init);
#else
module_init(cpufreq_gov_dbs_init);
#endif
module_exit(cpufreq_gov_dbs_exit);
